#!/usr/bin/python
# -*- coding: utf-8 -*-

import logging
import sys
import time
from socket import gethostname

from fluent import handler, sender

from drucker.utils import DruckerConfig
from .logger_interface import SystemLoggerInterface, ServiceLoggerInterface


class FluentSystemLogger(SystemLoggerInterface):

    def __init__(self, config: DruckerConfig = None) -> None:
        """
        Constructor
        :param config:
        """
        super().__init__()
        self.config = config
        logger_name = 'drucker'
        log_level = logging.NOTSET
        app_name = 'NONE'
        app_env = 'NONE'
        self.log = logging.getLogger(logger_name)
        self.log.setLevel(log_level)
        self.log.addHandler(self.__init_fluent_handler(app_name, app_env, log_level))
        if config is not None:
            self.init_app(config)

    def init_app(self, config: DruckerConfig):
        self.config = config
        app_name = config.APPLICATION_NAME
        app_env = config.SERVICE_LEVEL_ENUM.value
        log_level = logging.NOTSET
        self.log.addHandler(self.__init_fluent_handler(app_name, app_env, log_level))

    def __init_fluent_handler(self, app_name: str, app_env: str, log_level: int):
        custom_format = {
            'host': gethostname(),
            'short_message': '%(message)s',
            'timestamp': '%(created)d.%(msecs)d',
            'level': '%(loglevel)d',
            'service': 'drucker',
            'ml_service': app_name,
            'service_level': app_env
        }
        fluent_handler = handler.FluentHandler('drucker')
        formatter = handler.FluentRecordFormatter(custom_format)
        fluent_handler.setFormatter(formatter)
        fluent_handler.setLevel(log_level)
        return fluent_handler

    def exception(self, message: str) -> None:
        """
        emits exception to log
        :param message: error message
        """
        self.log.error(message, exc_info=sys.exc_info(), stack_info=True, extra={'loglevel': 3})

    def error(self, message: str) -> None:
        """
        emits error log
        :param message: log
        """
        self.log.error(message, extra={'loglevel': 3})

    def debug(self, message: str) -> None:
        """
        emits debug log
        :param message: log
        """
        self.log.debug(message, extra={'loglevel': 7})

    def info(self, message: str) -> None:
        """
        emits info log
        :param message: log
        """
        self.log.info(message, extra={'loglevel': 6})

    def warn(self, message: str) -> None:
        """
        emits warn log
        :param message: log
        """
        self.log.warning(message, extra={'loglevel': 4})


class FluentServiceLogger(ServiceLoggerInterface):

    def __init__(self, config: DruckerConfig = None):
        """
        Constructor
        :param config:
        """
        super().__init__()
        self.config = config
        app_name = 'NONE'
        app_env = 'NONE'
        self.logger = sender.FluentSender('drucker_service')
        self.ml_service = app_name
        self.service_level = app_env
        if config is not None:
            self.init_app(config)

    def init_app(self, config: DruckerConfig):
        self.config = config
        self.ml_service = config.APPLICATION_NAME
        self.service_level = config.SERVICE_LEVEL_ENUM.value

    def emit(self, request, response, suppress_log_inout: bool = False) -> None:
        """
        emits service log
        :param request:
        :param response:
        :param suppress_log_inout:
        :return:
        """
        try:
            if suppress_log_inout:
                ml_input = ''
                ml_output = ''
            else:
                ml_input = super().to_str_from_request(request)
                ml_output = super().to_str_from_response(response)

            self.logger.emit(None, {
                'host': gethostname(),
                'short_message': 'prediction result.',
                'timestamp': int(time.time() * 1000) / 1000,
                'level': logging.INFO,
                'service': 'drucker',
                'ml_service': self.ml_service,
                'service_level': self.service_level,
                'ml_input': ml_input,
                'ml_output': ml_output
            })
        except:
            try:
                FluentSystemLogger(self.config).exception("can't write log")
            except:
                pass
